@@ -15,14 +15,17 @@ import Control.Monad.Trans.State
 import Control.Monad.IO.Class
 import Data.ByteString (ByteString)
 import Data.Char
+import Data.Fixed (Centi, Pico, showFixed)
 import Data.Foldable (for_, traverse_)
 import Data.List (elemIndex)
 import Data.List.Split (chunksOf)
+import Data.IORef
 import Data.Maybe (fromMaybe)
 import Data.Monoid
 import Data.Set (Set)
 import Data.Text (Text)
 import Data.Time
+import Data.Time.Clock.POSIX (utcTimeToPOSIXSeconds)
 import Data.Traversable (for)
 import Graphics.Vty
 import Network.Connection
@@ -45,6 +48,7 @@ import Irc.Message
 import Irc.Model
 import Irc.RateLimit
 
+import ConnectCmds (connectCmds)
 import ClientState
 import Connection (connect, getRawIrcLine)
 import CommandArgs
@@ -86,28 +90,33 @@ main = do
                      recvChan settings hErr
        (width,height) <- displayBounds (outputIface vty)
        zone <- getCurrentTimeZone
-       driver vty vtyEventChan recvChan ClientState
-         { _clientServer0         = server0
-         , _clientRecvChan        = recvChan
-         , _clientErrors          = hErr
-         , _clientFocus           = ChannelFocus ""
-         , _clientDetailView      = False
-         , _clientTimeView        = True
-         , _clientMetaView        = True
-         , _clientEditBox         = Edit.empty
-         , _clientTabPattern      = Nothing
-         , _clientScrollPos       = 0
-         , _clientHeight          = height
-         , _clientWidth           = width
-         , _clientIgnores         = mempty
-         , _clientHighlights      = mempty
-         , _clientMessages        = mempty
-         , _clientNickColors      = defaultNickColors
-         , _clientAutomation      = [ctcpHandler,cancelDeopTimerOnDeop]
-         , _clientTimers          = mempty
-         , _clientTimeZone        = zone
-         , _clientConfig          = view cmdArgConfigValue args
-         }
+
+       let st0 = ClientState
+               { _clientServer0         = server0
+               , _clientRecvChan        = recvChan
+               , _clientErrors          = hErr
+               , _clientFocus           = ChannelFocus ""
+               , _clientDetailView      = False
+               , _clientTimeView        = True
+               , _clientMetaView        = True
+               , _clientFullView        = False
+               , _clientEditBox         = Edit.empty
+               , _clientTabPattern      = Nothing
+               , _clientScrollPos       = 0
+               , _clientHeight          = height
+               , _clientWidth           = width
+               , _clientIgnores         = mempty
+               , _clientHighlights      = mempty
+               , _clientMessages        = mempty
+               , _clientNickColors      = defaultNickColors
+               , _clientAutomation      = [ctcpHandler,cancelDeopTimerOnDeop,connectCmds]
+               , _clientTimers          = mempty
+               , _clientTimeZone        = zone
+               , _clientConfig          = view cmdArgConfigValue args
+               }
+       st1 <- schedulePing st0
+
+       driver vty vtyEventChan recvChan st1
 
 withVty :: (Vty -> IO a) -> IO a
 withVty k =
@@ -121,30 +130,50 @@ startIrcConnection ::
   Maybe Handle                   {- ^ error log              -} ->
   IO ClientConnection
 startIrcConnection config recvChan settings hErr =
-  do h            <- connect config settings
-     sendChan     <- atomically newTChan
-     sendThreadId <- forkIO (sendLoop sendChan h)
-     recvThreadId <- forkIO (socketLoop recvChan h hErr)
+  do connectRes <- try (connect config settings)
+     case connectRes of
+       Left e -> connectionFailed e
+       Right h -> connectionSuceeded h
 
-     let utf8Bytes = Text.encodeUtf8 . Text.pack
-         nick = mkId (utf8Bytes (view ssNick settings))
-         user = utf8Bytes (view ssUser settings)
-         real = utf8Bytes (view ssReal settings)
-         sasl = over (mapped.both) utf8Bytes (view ssSaslCredential settings)
-         pass = over mapped        utf8Bytes (view ssPassword settings)
+  where
+  utf8Bytes = Text.encodeUtf8 . Text.pack
+  nick = mkId (utf8Bytes (view ssNick settings))
+  user = utf8Bytes (view ssUser settings)
+  real = utf8Bytes (view ssReal settings)
+  sasl = over (mapped.both) utf8Bytes (view ssSaslCredential settings)
+  pass = over mapped        utf8Bytes (view ssPassword settings)
 
-     initializeConnection pass nick user real h
+  connectionSuceeded h =
+    do sendChan     <- atomically newTChan
+       connectedRef <- newIORef True
+       sendThreadId <- forkIO (sendLoop connectedRef sendChan h)
+       recvThreadId <- forkIO (socketLoop recvChan h hErr)
+       initializeConnection pass nick user real h
+       return ClientConnection
+         { _ccServerSettings = settings
+         , _ccConnection     = defaultIrcConnection
+                                 { _connNick = nick
+                                 , _connSasl = sasl
+                                 }
+         , _ccSendChan       = Just (connectedRef, sendChan)
+         , _ccRecvThread     = Just recvThreadId
+         , _ccSendThread     = Just sendThreadId
+         }
 
-     return ClientConnection
-       { _ccServerSettings = settings
-       , _ccSendChan       = sendChan
-       , _ccConnection     = defaultIrcConnection
-                               { _connNick = nick
-                               , _connSasl = sasl
-                               }
-       , _ccRecvThread     = recvThreadId
-       , _ccSendThread     = sendThreadId
-       }
+  connectionFailed (SomeException e) =
+    do let eUtf8 = utf8Bytes (show e)
+       now <- getCurrentTime
+       atomically (writeTChan recvChan (now, Error eUtf8))
+       return ClientConnection
+         { _ccServerSettings = settings
+         , _ccConnection     = defaultIrcConnection
+                                 { _connNick = nick
+                                 , _connSasl = sasl
+                                 }
+         , _ccSendChan       = Nothing
+         , _ccRecvThread     = Nothing
+         , _ccSendThread     = Nothing
+         }
 
 initializeConnection ::
   Maybe ByteString {- ^ server password -} ->
@@ -181,6 +210,11 @@ driver vty vtyEventChan ircMsgChan st0 =
 
   processTimerEvent e st =
     case e of
+      TransmitPing ->
+        do sendTimestampPing st
+           st' <- schedulePing st
+           considerTimers st'
+
       DropOperator chan ->
         do clientSend (modeCmd chan ["-o",views connNick idBytes conn]) st
            considerTimers st
@@ -242,14 +276,14 @@ interpretLogicOp _ (Record target message r) =
 -- Key Event Handlers!
 ------------------------------------------------------------------------
 
-data EventResult
+data KeyEventResult
   = KeepGoing ClientState
   | Exit
 
 changeInput :: (Edit.EditBox -> Edit.EditBox) -> ClientState -> ClientState
 changeInput f st = clearTabPattern (over clientEditBox f st)
 
-inputLogic :: ClientState -> (Image, IO EventResult)
+inputLogic :: ClientState -> (Image, IO KeyEventResult)
 inputLogic st =
   case clientInput st of
     '/':_ -> case commandEvent st of
@@ -257,13 +291,14 @@ inputLogic st =
                (img, Just m ) -> (img, m)
     txt -> (stringWithControls defAttr txt, fmap KeepGoing (doSendMessageCurrent SendPriv st))
 
-keyEvent :: Key -> [Modifier] -> ClientState -> IO EventResult
+keyEvent :: Key -> [Modifier] -> ClientState -> IO KeyEventResult
 keyEvent k ms st =
   let more = return . KeepGoing in
   case (k,ms) of
     (KFun 2   , []     ) -> more $ over clientDetailView not st
     (KFun 3   , []     ) -> more $ over clientTimeView   not st
     (KFun 4   , []     ) -> more $ over clientMetaView   not st
+    (KFun 5   , []     ) -> more $ over clientFullView   not st
     (KPageUp  , _      ) -> more $ scrollUp st
     (KPageDown, _      ) -> more $ scrollDown st
     (KChar 'n', [MCtrl]) -> more $ nextFocus st
@@ -367,7 +402,7 @@ doSendMessage sendType target message st =
 
   who = UserInfo (view connNick conn) Nothing Nothing
 
-commandEvent :: ClientState -> (Image, Maybe (IO EventResult))
+commandEvent :: ClientState -> (Image, Maybe (IO KeyEventResult))
 commandEvent st = commandsParser (clientInput st)
                         (exitCommand : normalCommands)
  where
@@ -555,22 +590,28 @@ commandEvent st = commandsParser (clientInput st)
 
   , ("reconnect", [], pure (doReconnect st'))
 
+  , ("ping", [],
+    (\msg -> st' <$ doPingCmd msg st')
+    <$> pRemainingNoSp)
+
   ]
 
--- Exploratory!
+doPingCmd :: String -> ClientState -> IO ()
+doPingCmd ""  st = sendTimestampPing st
+doPingCmd msg st = clientSend (pingCmd (Text.encodeUtf8 (Text.pack msg))) st
+
 doReconnect :: ClientState -> IO ClientState
 doReconnect st =
   do let server0 = view clientServer0 st
-     views ccRecvThread killThread server0
-     views ccSendThread killThread server0
+     traverse_ killThread (view ccRecvThread server0)
+     traverse_ killThread (view ccSendThread server0)
 
      let settings = view ccServerSettings server0
          recvChan = view clientRecvChan st
          hErr     = view clientErrors st
          config   = view clientConfig st
      server0' <- startIrcConnection config recvChan settings hErr
-     let st' = set clientServer0 server0' st
-     return st'
+     return (set clientServer0 server0' st)
 
 doNick ::
   ClientState ->
@@ -807,12 +848,24 @@ textbox st
 
 dividerImage :: ClientState -> Image
 dividerImage st
-  = extendToWidth (nickPart <|> channelPart)
+  = extendToWidth (nickPart <|> channelPart <|> lagPart)
   where
   conn = view (clientServer0 . ccConnection) st
 
   myNick = view connNick conn
 
+  lagPart =
+    case view connPingTime conn of
+      Just s ->
+       let s' = realToFrac s :: Centi -- truncate to two decimal places
+           sStr = showFixed True s' <> "s"
+       in
+                string defAttr "-["
+            <|> string (withForeColor defAttr yellow) sStr
+            <|> string defAttr "]"
+
+      _ -> emptyImage
+
   channelPart =
     ifoldr (\i x xs -> drawOne i x <|> xs)
            emptyImage
@@ -874,13 +927,15 @@ dividerImage st
 -- Event loops
 ------------------------------------------------------------------------
 
-sendLoop :: TChan ByteString -> Connection -> IO ()
-sendLoop queue h =
+sendLoop :: IORef Bool -> TChan ByteString -> Connection -> IO ()
+sendLoop connectedRef queue h =
   do r <- newRateLimitDefault
      forever $
        do x <- atomically (readTChan queue)
           tickRateLimit r
           connectionPut h x
+  `catch` \(SomeException _) ->
+      writeIORef connectedRef False -- exit silently
 
 socketLoop :: TChan (UTCTime, MsgFromServer) -> Connection -> Maybe Handle -> IO ()
 socketLoop chan h hErr =
@@ -903,21 +958,33 @@ vtyEventLoop chan vty = forever (atomically . writeTChan chan =<< nextEvent vty)
 getOne :: Connection -> Maybe Handle -> IO (UTCTime, MsgFromServer)
 getOne h hErr =
     do xs <- getRawIrcLine h
-       let debug x = for_ hErr (`hPrint` x)
        case parseRawIrcMsg xs of
-         Nothing ->
-           do debug xs
-              t <- getCurrentTime
-              return (t,Error ("Parse: " <> xs))
-         Just msg ->
-           do t <- maybe getCurrentTime return (msgTime msg)
-              case ircMsgToServerMsg msg of
-                Nothing ->
-                  do debug msg
-                     return (t,Error ("Unknown: " <> xs))
-                Just x ->
-                  do debug x
-                     return (t,x)
+         _ | B.null xs -> connectionClosed
+         Nothing       -> unparsableLine xs
+         Just rawMsg   ->
+           do t <- maybe getCurrentTime return (msgTime rawMsg)
+              case ircMsgToServerMsg rawMsg of
+                Nothing -> unhandledMsg t rawMsg
+                Just msg -> handledMsg t msg
+  where
+  debugPrint x = for_ hErr (`hPrint` x)
+
+  connectionClosed =
+    do t <- getCurrentTime
+       return (t,Error "Connection closed")
+
+  unparsableLine xs =
+    do debugPrint xs
+       t <- getCurrentTime
+       return (t,Error ("Unparsable IRC line: " <> xs))
+
+  unhandledMsg t rawMsg =
+    do debugPrint rawMsg
+       return (t,Error ("Unhandled IRC line: " <> Text.encodeUtf8 (Text.pack (show rawMsg))))
+
+  handledMsg t msg =
+    do debugPrint msg
+       return (t,msg)
 
 readEitherTChan :: TChan a -> TChan b -> IO (Either a b)
 readEitherTChan a b =
@@ -981,3 +1048,17 @@ defaultNickColors :: [Color]
 defaultNickColors =
   [cyan, magenta, green, yellow, blue,
    brightCyan, brightMagenta, brightGreen, brightBlue]
+
+schedulePing :: ClientState -> IO ClientState
+schedulePing st =
+  do now <- getCurrentTime
+     let pingTime = addUTCTime 60 now
+     return (addTimerEvent pingTime TransmitPing st)
+
+sendTimestampPing :: ClientState -> IO ()
+sendTimestampPing st =
+  do now <- getCurrentTime
+     let ts = realToFrac (utcTimeToPOSIXSeconds now) :: Pico
+         chopTrailingZeros = True
+         tsStr = B8.pack (showFixed chopTrailingZeros ts)
+     clientSend (pingCmd tsStr) st
